//
// Program for lecturing purposes by Marek Vinkler (xvinkl@fi.muni.cz).
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//


#include <cmath>
#include <string>

#include "shaderDemo.h"

// Include GLEW to access OpenGL 3.3 functions
#include <GL/glew.h>

// GLUT is the toolkit to interface with the OS
#include <GL/freeglut.h>

//----------------------------------------------------------------
// TODO:
// 1) Get location of the "position" attribute.
// 2) Create VAO containing the triangle geometry.
// 3) Use the programe and render the VAO.
// 4) Get location of the "m_pvm" uniform.
// 5) Set the "m_pvm" uniform with projection matrix.
// 6) Set fragment data location for varible "outputF".
// 7) Rotate the projection matrix.

//----------------------------------------------------------------

// Projection matrix
float projMatrix[16];

// Shaders and programs
GLuint v, f, p;

// Vertex attribute
GLint VERTEX_COORD_ATTRIB;

// Uniforms
GLint UNIFORM_MODELVIEWPROJ;

// Vao
GLuint vao;


// ------------------------------------------------------------
//
// Render stuff
//

void renderScene(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// TODO: Use our program
	
	// TODO: Set uniforms
	
	// TODO: Render
	
	// Swap buffers
	glutSwapBuffers();
}


// ------------------------------------------------------------
//
// Events from the Keyboard
//

void processKeys(unsigned char key, int xx, int yy)
{
	switch(key)
	{
		case 27:
			glutLeaveMainLoop();
			break;
	}

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// --------------------------------------------------------
//
// Aux func
//

void multMatrix4f(float *a, float *b, float *c)
{
	for(int i = 0; i < 4; ++i)
	{
		for(int j = 0; j < 4; ++j)
		{
			c[j*4 + i] = 0.0f;
			for(int k = 0; k < 4; ++k)
			{
				c[j*4 + i] += a[k*4 + i] * b[j*4 + k]; 
			}
		}
	}
}

// --------------------------------------------------------

char* textFileRead(std::string fileName)
{

	FILE *fp;
	char *content = NULL;

	int count=0;

	if(fileName != "")
	{
		fp = fopen(fileName.c_str(),"rt");

		if (fp != NULL)
		{
			fseek(fp, 0, SEEK_END);
			count = ftell(fp);
			rewind(fp);

			if(count > 0)
			{
				content = (char *)malloc(sizeof(char) * (count+1));
				count = fread(content,sizeof(char),count,fp);
				content[count] = '\0';
			}
			fclose(fp);
		}
	}
	return content;
}

// --------------------------------------------------------

void printShaderInfoLog(GLuint obj)
{
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

    glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if(infologLength > 0)
    {
        infoLog = (char *)malloc(infologLength);
        glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
				printf("%s\n",infoLog);
        free(infoLog);
    }
}

// --------------------------------------------------------

void printProgramInfoLog(GLuint obj)
{
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

    glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if(infologLength > 0)
    {
        infoLog = (char *)malloc(infologLength);
        glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
				printf("%s\n",infoLog);
        free(infoLog);
    }
}


// --------------------------------------------------------
//
// Shader Stuff
//

GLuint setupShaders()
{
	// Shader for models
	char *vs, *fs;

	// Setup
	v = glCreateShader(GL_VERTEX_SHADER);
	f = glCreateShader(GL_FRAGMENT_SHADER);	

	vs = textFileRead("shaders/helloWorld.vert");
	fs = textFileRead("shaders/helloWorld.frag");

	const char * vv = vs;
	const char * ff = fs;

	glShaderSource(v, 1, &vv, NULL);
	glShaderSource(f, 1, &ff, NULL);

	free(vs);
	free(fs);

	glCompileShader(v);
	glCompileShader(f);

	p = glCreateProgram();

	glAttachShader(p, v);
	glAttachShader(p, f);

	glLinkProgram(p);

	// TODO: Get locations of vertex attributes and uniforms
	// Vertex attributes
	
	// Uniforms
	
	// Define projection matrix
	float f = 1.0f / tan (60.0f * (3.14f / 360.0f));
	float nearp = 0.1f;
	float farp = 1000.f;
	float ratio = 512.f / 512.f;

	memset(projMatrix, 0, 16*sizeof(float));
	projMatrix[0 * 4 + 0] = f / ratio;
	projMatrix[1 * 4 + 1] = f;
	projMatrix[2 * 4 + 2] = (farp + nearp) / (nearp - farp);
	projMatrix[3 * 4 + 2] = (2.0f * farp * nearp) / (nearp - farp);
	projMatrix[2 * 4 + 3] = -1.0f;
	projMatrix[3 * 4 + 3] = 0.0f;

	// TODO: Set fragment data location

	// Print logs
	printShaderInfoLog(v);
	printShaderInfoLog(f);
	printProgramInfoLog(p);
	
	return GL_TRUE; // Here is hoping :-)
}


// ------------------------------------------------------------
//
// Model loading and OpenGL setup
//

void initOpenGL()
{
	// Some GL settings
	glEnable(GL_DEPTH_TEST);
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

	// TODO: Setup vertex attributes
}


// ------------------------------------------------------------
//
// Main function
//

int main(int argc, char **argv)
{
//  GLUT initialization
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGBA);

	glutInitContextVersion(3, 3);
	glutInitContextProfile(GLUT_CORE_PROFILE);
	glutInitContextFlags(GLUT_DEBUG);

	glutInitWindowPosition(100,100);
	glutInitWindowSize(512,512);
	glutCreateWindow("PV227 - Triangle");

//  Callback Registration
	glutDisplayFunc(renderScene);
	glutIdleFunc(renderScene);

//	Mouse and Keyboard Callbacks
	glutKeyboardFunc(processKeys);

//	Return from main loop
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

//	Init GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	printf("Vendor: %s\n", glGetString(GL_VENDOR));
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("Version: %s\n", glGetString(GL_VERSION));
	printf("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));

	if(!setupShaders())
		return(1);

	initOpenGL();

//  GLUT main loop
	glutMainLoop();

// Clean up
	glDetachShader(p, v);
	glDetachShader(p, f);
	
	glDeleteShader(v);
	glDeleteShader(f);
	
	glUseProgram(0);
	glDeleteProgram(p);

	return(0);
}

